/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	swixcvt.h
//
#pragma once

#include <SWI-Prolog.h>
#include <stdio.h>
#include <stdlib.h>
#include <map>
#include "declcom.h"
#include "basics.h"
#include "zstr.h"

///////////////////////////////////////////////////////////////////////
//	Interface smart pointer declarations
///////////////////////////////////////////////////////////////////////

//  Define IP_IDispatch
DECL_IP_U(IDispatch,IID_IDispatch);
DECL_IP_U(ITypeComp,IID_ITypeComp);
DECL_IP_U(ITypeInfo,IID_ITypeInfo);
DECL_IP_U(ITypeInfo2,IID_ITypeInfo2);


typedef vector<LONG> VLONG;

///////////////////////////////////////////////////////////////////////
//	ActiveX object registration
///////////////////////////////////////////////////////////////////////

typedef int ACTXTOKEN;
typedef vector<ACTXTOKEN> VATKN;


class ACTXITEM 
{
  protected:
	ACTXTOKEN _atkn;
	IP_IDispatch _ipId;
	int _iContext;

  public:

	ACTXITEM ()
		: _atkn(-1),
		_iContext(-1)
		{}

	ACTXITEM ( ACTXTOKEN atkn, IP_IDispatch & ipId, int iContext )
		: _atkn(atkn),
		_ipId( ipId ),
		_iContext(iContext)
		{}

	IP_IDispatch & IPId ()
		{ return _ipId; }

	ACTXTOKEN Atkn () const 
		{ return _atkn; }

	int & IContext ()
		{ return _iContext; }
};

typedef map<ACTXTOKEN,ACTXITEM> MPTKNIP;


///////////////////////////////////////////////////////////////////////
//	The ActiveX "Journal" (registration) class
///////////////////////////////////////////////////////////////////////

struct ACTXJOURNAL
{
  protected:
	//  Next token to be assigned
	ACTXTOKEN _tknNext;
	//  Map between tokens and IDispatch pointers
	MPTKNIP _mptknip;
	//  Context depth indicator
	int _iContext;
	
  public:
	ACTXJOURNAL ()
		: _tknNext(0),
		_iContext(-1)
		{}

	ACTXTOKEN nextToken ()
	{
		return _tknNext++;
	}

	//  Bump the "context" depth
	int IPushContext ()
	{
		return ++_iContext;
	}

	//  Remove all objects created on and below the context given
	void PopContext ( int iContext )
	{
		VATKN vatkn;

		for ( MPTKNIP::iterator itmp = _mptknip.begin();
				itmp != _mptknip.end();
				itmp++ )
		{
			ACTXITEM & aitem = (*itmp).second;
			if ( aitem.IContext() >= iContext )
				vatkn.push_back( aitem.Atkn() );
		}				

		for ( int itkn = 0; itkn < vatkn.size(); itkn++ )
		{
			erase( vatkn[itkn] );
		}
		_iContext = iContext - 1;
	}

	//  Store an IDispatch i/f reference.  Assign it a token and a context.
	ACTXTOKEN registerObject ( IP_IDispatch & ipIf, ACTXTOKEN * patkn = NULL )
	{
		ACTXTOKEN atkn = patkn 
					   ? *patkn
					   : nextToken();
		_mptknip[atkn] = ACTXITEM(atkn, ipIf, _iContext );

		return atkn;
	}

	//  Register an object which may already be registered.
	ACTXTOKEN registerDispatch ( IP_IDispatch & ipId )
	{
		for ( MPTKNIP::iterator itmp = _mptknip.begin();
				itmp != _mptknip.end();
				itmp++ )
		{
			ACTXITEM & aitem = (*itmp).second;
			if ( aitem.IPId() == ipId )
				return aitem.Atkn();
		}				
		return registerObject( ipId );
	}

	//  Destroy all registered objects.
	void clear ()
	{
		_mptknip.clear();
		_tknNext = 0;
		_iContext = -1;
	}

	//  Return true if the item is registered
	bool bRegistered ( ACTXTOKEN atkn )
	{
		MPTKNIP::iterator itmp = _mptknip.find( atkn );
		return itmp != _mptknip.end();
	}

	//  Remove the item from the map
	void erase ( ACTXTOKEN atkn )
	{
		MPTKNIP::iterator itmp = _mptknip.find( atkn );
		if ( itmp != _mptknip.end() )
			_mptknip.erase( itmp );
	}

	//  Locate an item by token; return the dispatch pointer
	bool bFind ( ACTXTOKEN atkn, IP_IDispatch & ipId )
	{
		ipId = NULL;
		MPTKNIP::iterator itmp = _mptknip.find( atkn );
		if ( itmp == _mptknip.end() )
			return false;
		ipId = (*itmp).second.IPId();
		return true;
	}

	ACTXITEM * PActxItem ( ACTXTOKEN atkn )
	{
		MPTKNIP::iterator itmp = _mptknip.find( atkn );
		if ( itmp == _mptknip.end() )
			return NULL;
		return & (*itmp).second;
	}

	//  Force the given item into the global context or local context
	bool bSetContext ( ACTXTOKEN atkn, bool bGlobal = true )
	{
		MPTKNIP::iterator itmp = _mptknip.find( atkn );
		if ( itmp == _mptknip.end() )
			return false;
		ACTXITEM & aitem = (*itmp).second;
		aitem.IContext() = bGlobal 
						  ? -1 
						  : _iContext;
		return true;
	}
};


//
//	Class Id and guid conversion functions
//
extern bool bStringFromGuid ( const GUID & guid, ZSTR & zsGuid );
extern bool bGuidFromString ( TSZC tszcGuid, GUID & guid );
extern bool bGetClsId ( TSZC tszcProgId, CLSID & clsid );
extern bool bGetClsIdString ( TSZC tszcProgId, ZSTR & zsClsId );

//	
//	VARIANT <-> Prolog term conversions
//

//  Convert a VARIANT to an integer; supports VT_{U}I{2,4,INT}, among others
bool bConvertVariantToInt ( _variant_t & vnt, int & iResult );

//	Scan a list in order to see if its elements are compatible; returns
//		the count of items in the list and their VARTYPE.
//
//		If input VARTYPE is VT_EMPTY, it will be set to the common element type,
//			and all elements must be of that type.
//		If input VARTYPE is VT_VARIANT, any element type is allowed.
//		If input VARTYPE is other, all elements must be of that type.
//
bool bPrepareList ( term_t tList, VARTYPE & vt, int & citems );

//
//  Validate that a term is a SAFEARRAY wrapper funtor, 
//		which looks like:
//
//			activeXSafearray(VariantType,ListOfElements).
//
bool bValidActiveXSafearray ( term_t tSafearray, 
							VARTYPE & vt,
							term_t & tList );

//
//	Determine if the given term is a valid functor for our ActiveX layer.
//  	In Prolog, they look like "activeXObject(n)", where n is the index in
//     	the "actxJournal" map.
//
bool bValidActiveXReference ( term_t tIDispatchFunctor, 
							IP_IDispatch & ipId, 
							ACTXTOKEN * piObjId = 0 );

//
//  Validate that a term is a VARIANT wrapper funtor, 
//		which looks like:
//
//			activeXVariant(VariantType,DataTerm).
//
bool bValidActiveXVariant ( term_t tVariant, 
							VARTYPE & vt,
							term_t & tData);

//
//	Convert a dimension of a SAFEARRAY into a Prolog list.
//	If this is not the lowest dimension, create a list of lists.
//
bool bConvertSafeArrayDimensionToList ( VARTYPE vt,
										SAFEARRAY * psfa,
										VLONG & vidim,
										int iDim,
										term_t tList );

//
//	Unify a single element of a SAFEARRAY with a term.
//		Element is represented by VARTYPE and pointer to data.
//
bool bConvertSafeArrayElementToTerm ( VARTYPE vt,
										void * pv,
										term_t tElem );


//
//	Unify an array variant with a list.   
//
bool bUnifyVariantArrayWithTerm ( 
	_variant_t & vnt,       		//  VARIANT to convert
	term_t & tTerm, 				//  Term to unify with 
	ACTXTOKEN * patkn = NULL );      //  Optional token for interface registration;

//
//	Create the special SAFEARRAY descriptor functor along with its list
//		and unify it with the given term.
//
bool bUnifyVariantArrayWithSafearrayFunctor ( 
	_variant_t & vnt, 
	term_t & tTerm,
	ACTXTOKEN * patkn = NULL );      //  Optional token for interface registration

//
//	Convert a returned VARIANT to a Prolog term.  
//
bool bUnifyVariantWithTerm ( 
	_variant_t & vnt,       			//  VARIANT to convert
	term_t & tTerm, 					//  Term to unify with 
	ACTXTOKEN * patkn = NULL );      	//  Optional token for interface registration

//
//	Create the special VARIANT functor used for unusual data types
//		and unify it with the given term.
//

bool bUnifyVariantWithVariantFunctor (
	_variant_t & vnt, 
	term_t & tTerm,
	ACTXTOKEN * patkn = NULL );		     //  Optional token for interface registration

//
//	Convert the given Prolog term to the desired VARIANT type.  
//     	Coerce special forms like "true", "false" and "optional" to 
//     	their correct types.
//
bool bConvertTermToVariant ( int vt, term_t tArg, _variant_t & vnt );

//
//      Using the ITypeInfo interface if necessary, return the
//      required VARTYPE for the TYPEDESC given.  This is normally
//      trivial, but may required dereferencing USERDEFINED types
//      through aliases.  ENUMs are the only supported user-defined types.
//
VARTYPE vtGetVarType ( IP_ITypeInfo & ipInfo, TYPEDESC & tdesc );

//
//	Convert a list of Prolog terms to a SAFEARRAY nestled in a VARIANT.
//	We do not allow unbound terms in a list.  All elements of a list must
//	be of the same type unless the VARTYPE is VT_VARIANT.
//
bool bConvertListToVariantSafearray ( term_t tList, _variant_t & vnt, VARTYPE vt = VT_EMPTY );

//  Convert an activeXVariant functor to a VARIANT.
bool bConvertVariantFunctorToVariant ( term_t tData, _variant_t & vnt, VARTYPE vt = VT_EMPTY );

//
//	Create a list containing the integer terms of the given date,
//		then attempt to unify the result with the given term.
//
bool bUnifyDateWithList ( const DATE & date, term_t tList );

//  Convert a list of integers to a DATE.
bool bConvertListToDate ( term_t tList, DATE & date );

//  Extract a double from an integer or float term
bool bGetDblFromTerm ( term_t tTerm, double & dbl );

//  Extract an integer from an integer or float term; float must
//		be integral.
bool bGetIntFromTerm ( term_t tTerm, int & i );

//	Determine if a double value is integral.
bool bIntegralFloat ( const double & dbl );
